import pprint
import yaml
import json
from flask import Flask, request, jsonify
from klue.swagger.spec import ApiSpec
from klue.swagger.server import spawn_server_api
from klue.swagger.api import default_error_callback
from klue.exceptions import KlueException
from mock import patch


#
# A base swagger YAML file
#

yaml_base = """
swagger: '2.0'
info:
  title: test
  version: '0.0.1'
  description: Just a test
host: pnt-login.elasticbeanstalk.com
schemes:
  - http
basePath: /v1
produces:
  - application/json
definitions:
  SessionToken:
    type: object
    description: An authenticated user''s session token
    properties:
      token:
        type: string
        description: Session token.
  Credentials:
    type: object
    description: A user''s login credentials.
    properties:
      email:
        type: string
        description: User''s email.
      int:
        type: string
        description: MD5 of user''s password, truncated to 16 first hexadecimal characters.
    required:
      - email
"""

def gen_app(ystr):
    swagger_dict = yaml.load(ystr)
    spec = ApiSpec(swagger_dict)
    app = Flask('test')
    spawn_server_api('somename', app, spec, default_error_callback, None)
    return app, spec

def assert_ok_reply(r, token):
    assert r.status_code == 200, r.status_code
    j = json.loads(r.data.decode("utf-8"))
    print("json reply: " + pprint.pformat(j))
    assert j['token'] == token

def assert_error(r, status, content):
    assert str(r.status_code) == str(status), "status [%s] == [%s]" % (r.status_code, status)
    assert content in r.status, "[%s] in [%s]" % (content, r.status)

#
# TESTS
#

yaml_no_param = yaml_base + """
paths:
  /v1/no/param:
    get:
      summary: blabla
      description: blabla
      produces:
        - application/json
      x-bind-server: klue.test.return_token
      x-auth-required: false
      responses:
        200:
          description: A session token
          schema:
            $ref: '#/definitions/SessionToken'
"""


@patch('klue.test.return_token')
def test_swagger_server_no_param(func):
    app, spec = gen_app(yaml_no_param)

    func.__name__ = 'return_token'
    SessionToken = spec.definitions['SessionToken']
    func.return_value = SessionToken(token='123')

    with app.test_client() as c:
        r = c.get('/v1/no/param')
        assert_ok_reply(r, '123')
        func.assert_called_once_with()


@patch('klue.test.return_token')
def test_swagger_server_no_result(func):
    app, spec = gen_app(yaml_no_param)

    func.__name__ = 'return_token'
    func.return_value = None

    with app.test_client() as c:
        r = c.get('/v1/no/param')
        assert_error(r, 500, 'INTERNAL SERVER ERROR')


@patch('klue.test.return_token')
def test_swagger_server_pass_through_responses(func):
    app, spec = gen_app(yaml_no_param)

    with app.test_request_context('/'):

        func.__name__ = 'return_token'
        r = jsonify({'foo': 'bar'})
        r.status_code = 534
        func.return_value = r

        with app.test_client() as c:
            r = c.get('/v1/no/param')
            print("sc: %s" % str(r.status_code))
            assert r.status_code == 534, r.status_code
            j = json.loads(r.data.decode("utf-8"))
            print("json reply: " + pprint.pformat(j))
            assert j['foo'] == 'bar'


@patch('klue.test.return_token')
def test_swagger_invalid_server_return_value(func):
    app, spec = gen_app(yaml_no_param)

    func.__name__ = 'return_token'
    func.return_value = {'a': 1}

    with app.test_client() as c:
        r = c.get('/v1/no/param')
        assert_error(r, 500, 'INTERNAL SERVER ERROR')
#         try:
#             c.get('/v1/no/param')
#         except Exception as e:
#             print("GOT EXCEPT %s" % str(e))
#             assert_error(e, 'InternalServerError', 'Contains')
#         else:
#             assert 0

# TODO: enable this test when server-side validation is enabled
#
#    c = spec.definitions['Credentials']
#    func.return_value = c(email='asdasd')
#
#     with app.test_client() as c:
#         r = c.get('/v1/no/param')
#         print("sc: %s" % str(r.status_code))
#         assert r.status_code == 500, r.status_code
#         j = json.loads(r.data.decode("utf-8"))
#         print("json reply: " + pprint.pformat(j))
#         assert j['error'] == 'SERVER_ERROR'
#         assert 'did not return a class' in j['error_description']


yaml_in_body = yaml_base + """
paths:
  /v1/in/body:
    get:
      summary: blabla
      description: blabla
      parameters:
        - in: body
          name: whatever
          description: User login credentials.
          required: true
          schema:
            $ref: '#/definitions/Credentials'
      produces:
        - application/json
      x-bind-server: klue.test.return_token
      x-auth-required: false
      responses:
        200:
          description: A session token
          schema:
            $ref: '#/definitions/SessionToken'
"""

@patch('klue.test.return_token')
def test_swagger_server_param_in_body(func):
    app, spec = gen_app(yaml_in_body)

    func.__name__ = 'return_token'
    SessionToken = spec.definitions['SessionToken']
    Credentials = spec.definitions['Credentials']
    func.return_value = SessionToken(token='456')

    with app.test_client() as c:
        r = c.get('/v1/in/body', data=json.dumps({
            'email': 'a@a.a',
            'int': '123123',
        }))
        assert_ok_reply(r, '456')
        func.assert_called_once_with(Credentials(email='a@a.a', int='123123'))


yaml_in_query = yaml_base + """
paths:
  /v1/in/query:
    get:
      summary: blabla
      description: blabla
      parameters:
        - in: query
          name: foo
          description: foooo
          required: true
          type: string
        - in: query
          name: bar
          description: baaar
          required: true
          type: string
      produces:
        - application/json
      x-bind-server: klue.test.return_token
      x-auth-required: false
      responses:
        200:
          description: A session token
          schema:
            $ref: '#/definitions/SessionToken'
"""

@patch('klue.test.return_token')
def test_swagger_server_param_in_query(func):
    app, spec = gen_app(yaml_in_query)

    func.__name__ = 'return_token'
    SessionToken = spec.definitions['SessionToken']
    func.return_value = SessionToken(token='456')

    with app.test_client() as c:
        r = c.get('/v1/in/query?foo=aaaa&bar=bbbb')
        assert_ok_reply(r, '456')
        func.assert_called_once_with(bar='bbbb', foo='aaaa')


@patch('klue.test.return_token')
def test_swagger_server_param_in_query__missing_required_param(func):
    app, spec = gen_app(yaml_in_query)

    func.__name__ = 'return_token'
    SessionToken = spec.definitions['SessionToken']
    func.return_value = SessionToken(token='456')

    with app.test_client() as c:
        r = c.get('/v1/in/query?bar=bbbb')
        assert_error(r, 400, 'BAD REQUEST')
        func.assert_not_called()



def test_swagger_server_auth():
    pass


@patch('klue.test.return_token')
def test_unmarshal_request_error__missing_required_argument(func):
    app, spec = gen_app(yaml_in_body)

    func.__name__ = 'return_token'
    SessionToken = spec.definitions['SessionToken']
    Credentials = spec.definitions['Credentials']
    func.return_value = SessionToken(token='456')

    with app.test_client() as c:
        r = c.get('/v1/in/body', data=json.dumps({'bazzoom': 'thiswontwork'}))
        assert_error(r, 400, 'BAD REQUEST')
        func.assert_not_called()


@patch('klue.test.return_token')
def test_unmarshal_request_error__wrong_argument_format(func):
    app, spec = gen_app(yaml_in_body)

    func.__name__ = 'return_token'
    SessionToken = spec.definitions['SessionToken']
    Credentials = spec.definitions['Credentials']
    func.return_value = SessionToken(token='456')

    with app.test_client() as c:
        data = json.dumps({
            'email': 'a@2a.a',
            'int': [1, 2, 3],
        })

        r = c.get('/v1/in/body', data=data)

        # verify server reply
# TODO: uncomment when bravado-core start returning error on this
#         assert r.status_code == 400, r.status_code
#         j = json.loads(r.data.decode("utf-8"))
#         print("json reply: " + pprint.pformat(j))
#         assert j['error'] == 'INVALID_PARAMETER'
#         assert j['status'] == 400
#         assert "email' is a required property" in j['error_description']


yaml_in_path = yaml_base + """
paths:
  /v1/in/{item}/foo/{path}:
    get:
      summary: blabla
      description: blabla
      parameters:
        - in: path
          name: item
          description: foooo
          required: true
          type: string
        - in: path
          name: path
          description: baaar
          required: true
          type: string
      produces:
        - application/json
      x-bind-server: klue.test.return_token
      x-auth-required: false
      responses:
        200:
          description: A session token
          schema:
            $ref: '#/definitions/SessionToken'
"""

@patch('klue.test.return_token')
def test_swagger_server_param_in_path(func):
    app, spec = gen_app(yaml_in_path)

    func.__name__ = 'return_token'
    SessionToken = spec.definitions['SessionToken']
    func.return_value = SessionToken(token='456')

    with app.test_client() as c:
        r = c.get('/v1/in/1234/foo/bob234')
        assert_ok_reply(r, '456')
        func.assert_called_once_with(item='1234', path='bob234')


# @patch('klue.test.return_token')
# def test_swagger_server_param_in_path__missing_required_param(func):
#     app, spec = gen_app(yaml_in_query)

#     func.__name__ = 'return_token'
#     SessionToken = spec.definitions['SessionToken']
#     func.return_value = SessionToken(token='456')

#     with app.test_client() as c:
#         r = c.get('/v1/in/query?bar=bbbb')
#         assert_error(r, 400, 'BAD REQUEST')
#         func.assert_not_called()



import pprint
import yaml
import responses
from httplib import HTTPResponse
from mock import patch, MagicMock
from klue.swagger.api import default_error_callback
from klue.swagger.spec import ApiSpec
from klue.swagger.client import generate_client_callers, _format_flask_url
from klue.exceptions import KlueException, ValidationError

def _slurp_yaml(yaml_str):
    swagger_dict = yaml.load(yaml_str)
    spec = ApiSpec(swagger_dict)
    callers_dict = generate_client_callers(spec, 10, default_error_callback)

    assert len(callers_dict.keys()) == 1
    assert 'do_test' in callers_dict

    handler = callers_dict['do_test']
    assert type(handler).__name__ == 'function'

    return handler, spec


yaml_query_param = """
swagger: '2.0'
info:
  title: test
  version: '0.0.1'
  description: Just a test
host: some.server.com
schemes:
  - http
basePath: /v1
produces:
  - application/json
paths:
  /v1/some/path:
    get:
      summary: blabla
      description: blabla
      parameters:
        - in: query
          name: arg1
          description: foooo
          required: true
          type: string
        - in: query
          name: arg2
          description: foooo
          required: true
          type: string
      produces:
        - application/json
      x-bind-server: whatever
      x-bind-client: do_test
      x-auth-required: false
      responses:
        '200':
          description: result
          schema:
            $ref: '#/definitions/Result'

definitions:

  Result:
    type: object
    description: result
    properties:
      foo:
        type: string
        description: blabla
      bar:
        type: string
        description: bloblo
"""

@responses.activate
def test_client_with_query_param():
    handler, _ = _slurp_yaml(yaml_query_param)

    responses.add(responses.GET, "http://some.server.com:80//v1/some/path",
                  body='{"foo": "a", "bar": "b"}', status=200,
                  content_type="application/json")

    res = handler(arg1='this', arg2='that').call()

    print("response: " + pprint.pformat(res))
    assert type(res).__name__ == 'Result'
    assert res.foo == 'a'
    assert res.bar == 'b'


@patch('klue.swagger.client.grequests')
def test_requests_parameters_with_query_param(grequests):
    grequests.get = MagicMock()
    handler, _ = _slurp_yaml(yaml_query_param)
    try:
        handler(arg1='this', arg2='that').call()
    except Exception as e:
        pass

    grequests.get.assert_called_once_with('http://some.server.com:80//v1/some/path',
                                          data=None,
                                          headers={'Content-Type': 'application/json'},
                                          params={'arg1': 'this', 'arg2': 'that'},
                                          timeout=10)


yaml_body_param = """
swagger: '2.0'
info:
  title: test
  version: '0.0.1'
  description: Just a test
host: some.server.com
schemes:
  - http
basePath: /v1
produces:
  - application/json
paths:
  /v1/some/path:
    post:
      summary: blabla
      description: blabla
      parameters:
        - in: body
          name: arg1
          description: foooo
          required: true
          schema:
            $ref: '#/definitions/Param'
      produces:
        - application/json
      x-bind-server: whatever
      x-bind-client: do_test
      x-auth-required: false
      responses:
        '200':
          description: result
          schema:
            $ref: '#/definitions/Result'

definitions:

  Result:
    type: object
    description: result
    properties:
      foo:
        type: string
        description: blabla
      bar:
        type: string
        description: bloblo

  Param:
    type: object
    description: param
    properties:
      arg1:
        type: string
        description: blabla
      arg2:
        type: string
        description: bloblo

"""

@responses.activate
def test_client_with_body_param():
    handler, spec = _slurp_yaml(yaml_body_param)

    responses.add(responses.POST, "http://some.server.com:80//v1/some/path",
                  body='{"foo": "a", "bar": "b"}', status=200,
                  content_type="application/json")

    # Only 1 parameter expected
    try:
        res = handler()
    except ValidationError as e:
        assert 1
    else:
        assert 0
    try:
        res = handler(1, 2)
    except ValidationError as e:
        assert 1
    else:
        assert 0

    # Send a valid parameter object
    model_class = spec.definitions['Param']
    param = model_class(arg1='a', arg2='b')

    res = handler(param).call()
    assert type(res).__name__ == 'Result'
    assert res.foo == 'a'
    assert res.bar == 'b'


@patch('klue.swagger.client.grequests')
def test_requests_parameters_with_body_param(grequests):
    handler, spec = _slurp_yaml(yaml_body_param)
    model_class = spec.definitions['Param']
    param = model_class(arg1='a', arg2='b')

    try:
        handler(param).call()
    except Exception as e:
        pass

    grequests.post.assert_called_once_with('http://some.server.com:80//v1/some/path',
                                           data='{"arg1": "a", "arg2": "b"}',
                                           headers={'Content-Type': 'application/json'},
                                           params=None,
                                           timeout=10)


def test_client_with_auth_required():
    pass


def test__format_flask_url():
    u = _format_flask_url(
        "/v1/seller/<item_id>/<path>/foo",
        {
            'item_id': '1234',
            'path': 'abcd',
        }
    )
    assert u == "/v1/seller/1234/abcd/foo"

    u = _format_flask_url(
        "/v1/seller/<item_id>/<path>/foo/<item_id>",
        {
            'item_id': 1234,
            'path': 'abcd',
        }
    )
    assert u == "/v1/seller/1234/abcd/foo/1234"


yaml_path_param = """
swagger: '2.0'
info:
  title: test
  version: '0.0.1'
  description: Just a test
host: some.server.com
schemes:
  - http
basePath: /v1
produces:
  - application/json
paths:
  /v1/some/{foo}/path/{bar}:
    get:
      summary: blabla
      description: blabla
      parameters:
        - in: path
          name: foo
          description: foooo
          required: true
          type: string
        - in: query
          name: bar
          description: baaar
          required: true
          type: string
      produces:
        - application/json
      x-bind-server: whatever
      x-bind-client: do_test
      x-auth-required: false
      responses:
        '200':
          description: result
          schema:
            $ref: '#/definitions/Result'

definitions:

  Result:
    type: object
    description: result
    properties:
      foo:
        type: string
        description: blabla
      bar:
        type: string
        description: bloblo
"""

@responses.activate
def test_client_with_path_param():
    handler, spec = _slurp_yaml(yaml_path_param)

    responses.add(responses.GET,
                  "http://some.server.com:80//v1/some/123/path/456",
                  body='{"foo": "a", "bar": "b"}',
                  status=200,
                  content_type="application/json")

    # Make a valid call
    res = handler(foo=123, bar=456).call()
    assert type(res).__name__ == 'Result'
    assert res.foo == 'a'
    assert res.bar == 'b'


@patch('klue.swagger.client.grequests')
def test_requests_parameters_with_path_params(grequests):
    handler, spec = _slurp_yaml(yaml_path_param)

    try:
        handler(foo=123, bar=456).call()
    except Exception as e:
        pass

    grequests.get.assert_called_once_with(
        'http://some.server.com:80//v1/some/123/path/456',
        data=None,
        headers={'Content-Type': 'application/json'},
        params=None,
        timeout=10)



import grequests
import pprint
import jsonschema
import json
import logging
import flask
from klue.exceptions import KlueException, ValidationError
from klue.utils import get_function
from bravado_core.response import unmarshal_response, OutgoingResponse


log = logging.getLogger(__name__)


try:
    from flask import _app_ctx_stack as stack
except ImportError:
    from flask import _request_ctx_stack as stack


def generate_client_callers(spec, timeout, error_callback):
    """Return a dict mapping method names to anonymous functions that
    will call the server's endpoint of the corresponding name as
    described in the api defined by the swagger dict and bravado spec"""

    callers_dict = {}

    def mycallback(endpoint):
        if not endpoint.handler_client:
            return

        log.info("Generating client for %s %s" % (endpoint.method, endpoint.path))

        callers_dict[endpoint.handler_client] = _generate_client_caller(spec, endpoint, timeout, error_callback)

    spec.call_on_each_endpoint(mycallback)

    return callers_dict



def _generate_client_caller(spec, endpoint, timeout, error_callback):

    url = "%s://%s:%s/%s" % (spec.protocol,
                             spec.host,
                             spec.port,
                             endpoint.path)

    decorator = None
    if endpoint.decorate_request:
        decorator = get_function(endpoint.decorate_request)

    method = endpoint.method.lower()
    if method not in ('get', 'post'):
        return error_callback(KlueException("BUG: method %s for %s is not supported. Only get and post are." %
                                            (endpoint.method, endpoint.path)))

    grequests_method = getattr(grequests, method)
    if decorator:
        grequests_method = decorator(grequests_method)
    import pprint

    def client(*args, **kwargs):
        """Call the server endpoint and handle marshaling/unmarshaling of parameters/result.

        client takes either a dict of query parameters, or an object representing the unique
        body parameter.
        """

        headers = {'Content-Type': 'application/json'}
        data = None
        params = None

        custom_url = url

        if hasattr(stack.top, 'call_id'):
            headers['KlueCallID'] = stack.top.call_id
        if hasattr(stack.top, 'call_path'):
            headers['KlueCallPath'] = stack.top.call_path

        if endpoint.param_in_query:
            # The query parameters are contained in **kwargs
            params = kwargs
            # TODO: validate params? or let the server do that...
        elif endpoint.param_in_body:
            # The body parameter is the first elem in *args
            if len(args) != 1:
                return error_callback(ValidationError("%s expects exactly 1 parameter" % endpoint.handler_client))
            data = json.dumps(spec.model_to_json(args[0]))
        elif endpoint.param_in_path:
            # Client expects arguments as a dict, not as a list
            assert len(args) == 0
            print("got url and params: " + pprint.pformat([url, kwargs]))
            custom_url = _format_flask_url(url, kwargs)

        # TODO: if request times-out, retry a few times, else return KlueTimeOutError
        # Call the right grequests method (get, post...)
        greq = grequests_method(custom_url,
                                data=data,
                                params=params,
                                headers=headers,
                                timeout=timeout)

        return ClientCaller(greq, endpoint.operation, endpoint.method, endpoint.path, error_callback)

    return client


def _format_flask_url(url, params):
    # TODO: make this code more robust: error if some params are left unmatched
    # or if url still contains placeholders after replacing
    for name, value in params.iteritems():
        url = url.replace("<%s>" % name, str(value))
    return url


class ClientCaller():

    def __init__(self, greq, operation, method, path, error_callback):
        self.operation = operation
        self.greq = greq
        self.method = method
        self.path = path
        self.error_callback = error_callback

    def call(self):
        # TODO: add retry handler to map
        log.info("Calling %s %s" % (self.method, self.path))
        responses = grequests.map([self.greq])
        assert len(responses) == 1
        response = responses[0]

        # If the remote-server returned an error, raise it as a local KlueException
        if str(response.status_code) != '200':
            log.warn("Call to %s %s returns error: %s" % (self.method, self.path, response.text))
            if 'error_description' in response.text:
                # We got a KlueException: unmarshal it and return as valid return value
                # UGLY FRAGILE CODE. To be replaced by proper exception scheme
                pass
            else:
                # Unknown exception...
                k = KlueException(response.text)
                k.status_code = response.status_code
                return self.error_callback(k)

        result = self._unmarshal(response)
        log.info("Call to %s %s returned an instance of %s" % (self.method, self.path, type(result)))
        return result

    def _unmarshal(self, response):
        # Now transform the request's Response object into an instance of a
        # swagger model
        try:
            result = unmarshal_response(response, self.operation)
        except jsonschema.exceptions.ValidationError as e:
            k = ValidationError(str(e))
            k.status_code = 500
            return self.error_callback(k)
        return result

def async_call(*client_callers):
    """Call these server endpoints asynchronously and return Model or Error objects"""
    # TODO: add retry handler to map
    responses = grequests.map(client_callers)
    results = []
    for i in xrange(1, len(responses)):
        client_caller = client_callers[i]
        response = responses[i]
        results.append(client_caller._unmarshal(response))
